#include <iostream>
#include <vector>
#include <set>
#include <algorithm>
#include "myvar.h"
#define MAX 10

// MAX 10 -> pow2s
const int pow2s[MAX] = {1,2,4,8,16,32,64,128,256,512};



typedef std::pair<
	std::pair<int, int>,
	bool
	> t_triple;


typedef std::set<t_triple> t_set_triple;

typedef std::vector<t_triple> t_vector_triple;

struct t_model {
	int n;
	int relation[MAX][MAX];
	int ortho_left[MAX];
	int ortho_right[MAX];

	t_set_triple t;	
	t_vector_triple v;	

	t_model() {}
	t_model (const int _n, const int (&_relation)[MAX][MAX], const int (&_ortho_left)[MAX], const int (&_ortho_right)[MAX], const t_set_triple &_t, const t_vector_triple &_v) {
		n = _n;
		for (int i = 0; i < _n; i++)
		for (int j = 0; j < _n; j++)
			relation[i][j] = _relation[i][j];

		for (int i = 0; i < n; i++) {
			ortho_left[i] = _ortho_left[i];
			ortho_right[i] = _ortho_right[i];
		}


		t = _t;
		v = _v;
	}
};

template <int N>
struct hyper_pseudo_EA {
	int op[N][N];

	void disp();
	int hyper_op(int set1, int set2);

	bool is_associative();
	bool has_unique_orthosupplement();
};

/* functions declarations */
void generate_all_relations_rec(int x, int y, int n, int(&relation)[MAX][MAX], std::vector<t_model> &models);
void generate_hyper_pseudo_EA3(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit = 1000000000); 
void generate_hyper_pseudo_EA4(const t_model &model, unsigned long int generated_limit = 1000000000);
void generate_hyper_pseudo_EA5(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit = 1000000000); 
void generate_hyper_pseudo_EA5_just(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit = 1000000000); 
void generate_hyper_pseudo_EA6(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit = 1000000000); 
void generate_hyper_pseudo_EA6_just(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit = 1000000000); 
/* */


int relation[MAX][MAX];

template <int N>
bool hyper_pseudo_EA<N>::has_unique_orthosupplement() {
	int 
		orthosupplements_left[N], orthosupplements_left_bitmap[N];

	for (int i = 0; i < N; i++) {
		orthosupplements_left[i] 
			= orthosupplements_left_bitmap[i] = 0;
	}

	for (int i = 0; i < N; i++) {
		bool found = false;

		for (int j = 0; j < N; j++) {
			if (op[j][i] & (1 << (N - 1))) {
				if (found)
					return false;
				found = true;
				if (!orthosupplements_left_bitmap[j]) {
					orthosupplements_left[j] = i;
					orthosupplements_left_bitmap[j] = 1;
				}
				else
					return false;
			}
		}

		if (!found)
			return false;
	}

	return true;
}

template <int N>
void hyper_pseudo_EA<N>::disp() {
	for (int i = 0; i < N; i++) {
		for (int j = 0; j < N; j++)
			std::cout << op[i][j] << ' ';
		std::cout << '\n';
	}
}

template <int N>
int hyper_pseudo_EA<N>::hyper_op(int set1, int set2) {
	int outcome = 0;
	for (int i = 0; i < N; i++) {
		if (set1 & (1 << i)) {
			for (int j = 0; j < N; j++) {
				if (set2 & (1 << j)) {
					if (op[i][j] == 0)
						return 0;

					outcome |= op[i][j];
				}
			}
		}
	}

	return outcome;
}

template <int N>
void generate_all_hyper_pseudo_EA(int x, int y, hyper_pseudo_EA<N> &hpea) {
	if (x == N) {
		if (hpea.is_associative() && hpea.has_unique_orthosupplement()) {
			hpea.disp();
			std::cout << "---\n";
		}
	}
	else {
		for (int i = 0; i < (1 << N); i++) {
			hpea.op[x][y] = i;
			if (y < N - 1)
				generate_all_hyper_pseudo_EA(x, y + 1, hpea);
			else
				generate_all_hyper_pseudo_EA(x + 1, 0, hpea);
		}

	}
}

template <int N>
bool hyper_pseudo_EA<N>::is_associative() {
	for (int i = 0; i < N; i++)
	for (int j = 0; j < N; j++) {
		if (op[i][j] != 0) {
			for (int k = 0; k < N; k++) {
				int res1 = hyper_op(op[i][j], 1<<k);

				if (res1 != 0) {
					if (op[j][k] == 0)
						return false;
					int res2 = hyper_op(1<<i, op[j][k]);
					if (res2 == 0)
						return false;
					if (res1 != res2)
						return false;
						
				}		
			}
		}
	}

	for (int j = 0; j < N; j++)
	for (int k = 0; k < N; k++) {
		if (op[j][k] != 0) {
			for (int i = 0; i < N; i++) {
				int res1 = hyper_op(1<<i, op[j][k]);

				if (res1 != 0) {
					if (op[i][j] == 0)
						return false;
					int res2 = hyper_op(op[i][j], 1<<k);
					if (res2 == 0)
						return false;

					if (res1 != res2)
						return false;
				}
			}
		}
	}

	return true;
}

int hyper_op(const int n, const int (&hop)[MAX][MAX], const int set1, const int set2) {
	int outcome = 0;
	for (int i = 0; i < n; i++) {
		if (set1 & (1 << i)) {
			for (int j = 0; j < n; j++) {
				if (set2 & (1 << j)) {
					if (hop[i][j] == 0)
						return 0;

					outcome |= hop[i][j];
				}
			}
		}
	}

	return outcome;
}


bool is_associative(const int n, const int (&hypea)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++) {
			if (hypea[i][j] != 0) {
				for (int k = 0; k < n; k++) {
					int res1 = hyper_op(n, hypea, hypea[i][j], 1 << k);

					if (res1 != 0) {
						if (hypea[j][k] == 0)
							return false;
						int res2 = hyper_op(n, hypea, 1 << i, hypea[j][k]);
						if (res2 == 0)
							return false;
						if (res1 != res2)
							return false;

					}
				}
			}
		}

	for (int j = 0; j < n; j++)
		for (int k = 0; k < n; k++) {
			if (hypea[j][k] != 0) {
				for (int i = 0; i < n; i++) {
					int res1 = hyper_op(n, hypea, 1 << i, hypea[j][k]);

					if (res1 != 0) {
						if (hypea[i][j] == 0)
							return false;
						int res2 = hyper_op(n, hypea, hypea[i][j], 1 << k);
						if (res2 == 0)
							return false;

						if (res1 != res2)
							return false;
					}
				}
			}
		}

	return true;
}




bool is_less_relation(int n, int(&rel1)[MAX][MAX], int(&rel2)[MAX][MAX]) {
	/*
        .789
	1.ab
	24.c
	356.
	*/


	// lower triangle

	int i = 1, j;
	while (i < n - 1) {
		j = i;
		while (j < n - 1) {
			j++;
			if (rel1[j][i] != rel2[j][i]) {
		//		std::cout << "i = " << i << ", j = " << j << '\n';
				if (rel1[j][i] < rel2[j][i])
					return true;
				else
					return false;
			}
		}
		i++;
	}

	// upper triangle
	i = 0;
	while (i < n - 1) {
		j = i;
		while (j < n - 1) {
			j++;
			if (rel1[i][j] != rel2[i][j]) {
		//		std::cout << "i = " << i << ", j = " << j << '\n';
				if (rel1[i][j] < rel2[i][j])
					return true;
				else
					return false;
			}
		}
		i++;
	}

	return false;
}
int rel1[MAX][MAX] = {
	{1, 0},
	{0, 1}
};

int rel2[MAX][MAX] = {
	{1, 1},
	{0, 1}
};

int rel3[MAX][MAX] = {
	{ 1, 0 },
	{ 1, 1 }
};

void generate_all_relations(int n) {
	int relation[MAX][MAX];
	
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++)
			relation[i][j] = 0;
	// reflexive
	for (int i = 0; i < n; i++)
		relation[i][i] = 1;

	
	std::vector<t_model> models;
	generate_all_relations_rec(0, 1, n, relation, models);
}

void generate_all_bounded_relations(int n, std::vector<t_model> &models) {
	int relation[MAX][MAX];
	
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++)
			relation[i][j] = 0;

	// reflexive
	for (int i = 0; i < n; i++)
		relation[i][i] = 1;

	// 0Rx
	for (int i = 1; i < n; i++)
		relation[0][i] = 1;

	// xR1

	for (int i = 0; i < n; i++)
		relation[i][n - 1] = 1;


	generate_all_relations_rec(1, 2, n, relation, models);
}
void get_next_element(int x, int y, int n, int(&relation)[MAX][MAX], std::vector<t_model> &models) {
	if (y < n - 2)
		generate_all_relations_rec(x, y + 1, n, relation, models);
	else
		generate_all_relations_rec(x + 1, x + 2, n, relation,models);
}

template <int N>
void swap(int(&perm)[N], int idx1, int idx2) {
	if (idx1 < N && idx2 < N && idx1 != idx2) {
		int tmp = perm[idx1];
		perm[idx1] = perm[idx2];
		perm[idx2] = tmp;
	}
}

template <int N>
bool next_perm(int idx_start, int idx_end, int(&perm2)[N]) {
	int i = idx_end - 1;
while (i >= idx_start && perm2[i] > perm2[i + 1])
i--;

int k = 1;
while (i + k < idx_end - k + 1) {
	swap(perm2, i + k, idx_end + 1 - k);
	k++;
}

if (i >= idx_start) {
	for (int k = i + 1; (k < N) && (k <= idx_end); k++) {
		if (perm2[i] < perm2[k]) {
			swap(perm2, i, k);
			return true;
		}
	}
}

return false;
}

void apply_isomorphism(int n, int(&relation)[MAX][MAX], int(&isomorphism)[MAX], int(&relation_iso)[MAX][MAX]) {
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++)
			relation_iso[isomorphism[i]][isomorphism[j]] = relation[i][j];

}
bool exists_isomorphic(int n, int(&relation)[MAX][MAX]) {
	// for each permutation 1..(n-2)
	//   compute_isomorphism -> ism
	//   if (ism < relation)
	//		return false;
	// return true;

	int perm[MAX];
	for (int i = 0; i < n; i++)
		perm[i] = i;

	// skip identity permutation
	while (next_perm(1, n - 2, perm)) {
		int rel_iso[MAX][MAX];

#if 0
		std::cout << "Perm: ";
		for (int i = 1; i <= n - 2; i++)
			std::cout << perm[i];
		std::cout << '\n';

		std::cout << "Rel    : ";
		for (int i = 0; i < n; i++)
			for (int j = 0; j < n; j++)
				std::cout << relation[i][j];
		std::cout << '\n';
#endif
		apply_isomorphism(n, relation, perm, rel_iso);

#if 0
		std::cout << "Rel ISO: ";
		for (int i = 0; i < n; i++)
			for (int j = 0; j < n; j++)
				std::cout << rel_iso[i][j];
		std::cout << '\n';
#endif


		if (is_less_relation(n, rel_iso, relation))
			return true;
	}

	return false;
}

bool admits_ortho_left(int n, int(&relation)[MAX][MAX], int(&ortho_left)[MAX]) {
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++) {
			if (relation[i][j] && !relation[ortho_left[j]][ortho_left[i]])
				return false;
			if (relation[ortho_left[j]][ortho_left[i]] && !relation[i][j])
				return false;
		}

	return true;
}

void get_right_perm(const int n, const int(&left_perm)[MAX], int(&right_perm)[MAX]) {
	for (int i = 0; i < n; i++)
		right_perm[left_perm[i]] = i;
}

typedef std::set<std::pair<int, int> > t_set_pair;
//bool operator==(const t_set_pair &s1, const t_set_pair &s2)

void disp_set_pair(const t_set_pair &tsp) {
	for (std::set<std::pair<int, int> >::iterator it = tsp.begin(); it != tsp.end(); ++it)
		std::cout << "(" << it->first << ", " << it->second << "), ";
	std::cout << '\n';
}

char get_elem_str(int n, int elem) {
	//std::cout << "\nELEM:" << elem << '\n';
	if (elem == 0)
		return '0';
	if (elem == n - 1)
		return '1';
	return char('a' + elem - 1);
}

void disp_setKKK(int n, int set) {
	std::cout << '{';
	bool first = true;
	for (int i = 0; i < n; i++) {
		if (set & (1 << i)) {
			if (first) {
				first = false;
			}
			else
				std::cout << ", ";
			std::cout << get_elem_str(n, i);
		}
	}
	std::cout << '}';
}

void disp_hypea(int n, int(&hypea)[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {
			disp_setKKK(n, hypea[i][j]);
			//std::cout << hypea[i][j];
			std::cout << ", ";
		}
		std::cout << '\n';
	}
}


//void generate_hyper_pseudo_EA2(const t_set_triple &it, int (&hypea)[MAX][MAX], const int n, const t_set_triple &plan, const int(&ortho_left)[MAX], const int(&ortho_right)[MAX], const int(&relation)[MAX][MAX]) {

//}
void generate_hyper_pseudo_EA_rec(t_set_triple::iterator it, int (&hypea)[MAX][MAX], const int n, const t_set_triple &plan, const int(&ortho_left)[MAX], const int(&ortho_right)[MAX], const int(&relation)[MAX][MAX]) {
	if (it == plan.end()) {
		if (is_associative(n, hypea)) {
			/*for (int i = 0; i < n; i++) {
				for (int j = 0; j < n; j++)
					std::cout << hypea[i][j] << ", ";
				std::cout << '\n';
			}*/
			disp_hypea(n, hypea);
			std::cout << "---\n";
		}
	}
	else {
		int x = (it->first).first, y = (it->first).second;
		bool b = it->second;
		//std::cout << "x = " << x << ", y = " << y << ", b = " << b << '\n';
		if (x == 0 && y == 0) {
			// {0}
			t_set_triple::iterator it_copy = it;
			it_copy++;
			hypea[x][y] = 1;
			generate_hyper_pseudo_EA_rec(it_copy, hypea, n, plan, ortho_left, ortho_right, relation);
			hypea[x][y] = 0;
		}
		else if (x == 0) {
			int end;
			if (b)
				end = 1 << n;
			else
				end = (1 << (n - 1)) + 1;
			
			for (int i = 1 << y; i < end; i++) {
				// y in 0 + y
				if (i & (1 << y)) {
					// for each t\in (0 + y), if y + x exists, t + x must exist
					bool found = false;
					{
						for (int xt = 0; (xt < n) && !found; xt++) {
							if (relation[y][ortho_right[xt]]) {
								for (int t = 0; t < n; t++)
									if (i & (1 << t) && !relation[t][ortho_right[xt]]) {
										found = true;
										break;
									}
							}
						}
					}
					if (!found) {
						t_set_triple::iterator it_copy = it;
						it_copy++;

						hypea[x][y] = i;
						generate_hyper_pseudo_EA_rec(it_copy, hypea, n, plan, ortho_left, ortho_right, relation);
						hypea[x][y] = 0;
					}
				}

			}
		}
		else if (y == 0) {

			int end;
			if (b)
				end = 1 << n;
			else
				end = (1 << (n - 1)) + 1;

			for (int i = 1 << x; i < end; i++) {
				// x in x + 0
				if (i & (1 << x)) {
					// for each t \in (x + 0), if y + x exists, y + t must exist
					bool found = false;
					{
						for (int yt = 0; (yt < n) && !found; yt++) {
							if (relation[yt][ortho_right[x]]) {
								for (int t = 0; t < n; t++)
									if (i & (1 << t) && !relation[yt][ortho_right[t]]) {
										found = true;
										break;
									}
							}
						}
					}

					if (!found) {
						t_set_triple::iterator it_copy = it;
						it_copy++;

						hypea[x][y] = i;
						generate_hyper_pseudo_EA_rec(it_copy, hypea, n, plan, ortho_left, ortho_right, relation);
						hypea[x][y] = 0;
					}
				}

			}
		}
		else {
			// ones
			if (b) {
				for (int i = 1 << (n - 1); i < 1 << n; i++) {
					hypea[x][y] = i;
					t_set_triple::iterator it_copy = it;
					it_copy++;
					generate_hyper_pseudo_EA_rec(it_copy, hypea, n, plan, ortho_left, ortho_right, relation);
					hypea[x][y] = 0;
				}
			}
			else {
				// others
				for (int i = 1; i < 1 << (n - 1); i++) {
					hypea[x][y] = i;
					t_set_triple::iterator it_copy = it;
					it_copy++;
					generate_hyper_pseudo_EA_rec(it_copy, hypea, n, plan, ortho_left, ortho_right, relation);
					hypea[x][y] = 0;
				}
			}
		}
	}
}

void generate_hyper_pseudo_EA(int n, const t_set_triple &plan, int(&ortho_left)[MAX], int(&ortho_right)[MAX], int(&relation)[MAX][MAX]) {
	int hypea[MAX][MAX];
	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++)
			hypea[i][j] = 0;
	
	generate_hyper_pseudo_EA_rec(plan.begin(), hypea, n, plan, ortho_left, ortho_right, relation);
}


void generate_all_relations_rec(int x, int y, int n, int(&relation)[MAX][MAX], std::vector<t_model> &models) {


//	std::cout << "x = " << x << '\n';
	if (x == n - 2) {
		if (!exists_isomorphic(n, relation)) {
			int ortho_left[MAX];
			for (int i = 1; i < n - 1 ; i++)
				ortho_left[i] = i;
			ortho_left[0] = n - 1;
			ortho_left[n - 1] = 0;

			std::vector<std::pair<t_set_pair, t_set_pair> > ALL;
			do {
				if (admits_ortho_left(n, relation, ortho_left)) {
					int ortho_right[MAX];
					get_right_perm(n, ortho_left, ortho_right);

					std::set<std::pair<int,int> > ones;
					for (int i = 0; i < n; i++)
						ones.insert(std::pair<int,int>(i, ortho_left[i]));

					std::set<std::pair<int,int> > oths;
					for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						if (i != j && relation[i][j]) {
							oths.insert(std::pair<int, int>(i, ortho_left[j]));
							oths.insert(std::pair<int, int>(ortho_right[j], i));
						}

					bool found = false;
					if (!ALL.empty()) 
					{
						// if for all possible isomorphisms
						// there is not the same (ones, oths) pair in ALL
						//	insert it into ALL
						int isomorphism[MAX];
						for (int i = 0; i < n; i++)
							isomorphism[i] = i;

						do {

							t_set_pair ones1, oths1;

							for (int i = 0; i < n; i++)
								ones1.insert(std::pair<int,int>(isomorphism[i], isomorphism[ortho_left[i]]));

							for (int i = 0; i < n; i++)
							for (int j = 0; j < n; j++)
								if (i != j && relation[i][j]) {
									oths1.insert(std::pair<int,int>(isomorphism[i], isomorphism[ortho_left[j]]));
									oths1.insert(std::pair<int, int>(isomorphism[ortho_right[j]], isomorphism[i]));
								}

#if 0
							std::cout << '\t';
							for (int k = 0; k < n; k++)
								std::cout << isomorphism[k];
							std::cout << '\n';
							
							std::cout << '\t';
							disp_set_pair(ones1);
							std::cout << '\t';
							disp_set_pair(oths1);
#endif
	
							std::vector<std::pair<t_set_pair, t_set_pair> >::iterator it;
							it = std::find(ALL.begin(), ALL.end(), std::pair<t_set_pair, t_set_pair>(ones1,oths1));

							if (it != ALL.end()) {
								found = true;
//								std::cout << "FOUND!\n";
								break;
							}
						} while (next_perm(1, n - 2, isomorphism));
					}



					if (!found) {

					//	std::vector<std::pair<t_set_pair, t_set_pair> >::iterator it;
					//	it = std::find(ALL.begin(), ALL.end(), std::pair<t_set_pair, t_set_pair>(ones,oths));
					//	if (it == ALL.end()) {
#if DEBUG
							for (int i = 0; i < n; i++) {
								for (int j = 0; j < n; j++) {
									std::cerr << relation[i][j];
								}
								std::cerr << '\n';
								
							}
							std::cerr << "-\n";

							std::cerr << "Ortho_left : ";
							for (int i = 0; i < n; i++)
								std::cerr << ortho_left[i];
							std::cerr << '\n';

							std::cerr << "Ortho_right: ";
							for (int i = 0; i < n; i++)
								std::cerr << ortho_right[i];
							std::cerr << '\n';


						std::cerr << "Ones: ";
						for (std::set<std::pair<int, int> >::iterator it = ones.begin(); it != ones.end(); ++it)
							std::cerr << "(" << it->first << ", " << it->second << "), ";
						std::cerr << '\n';

						std::cerr << "Oths: ";
						for (std::set<std::pair<int, int> >::iterator it = oths.begin(); it != oths.end(); ++it)
							std::cerr << "(" << it->first << ", " << it->second << "), ";
						std::cerr << '\n';
							std::cerr << "---\n";
						
#endif
							ALL.push_back(std::pair<t_set_pair, t_set_pair>(ones,oths));

							// merge ones, oths
				
							std::set<
								std::pair<
									std::pair<int, int>,
									bool
								> 
							> merge_set;

						/*	merge_set.insert(
								std::pair <
								std::pair<int, int>,
								bool
								>(std::pair<int, int>(0, 0), true));*/

							
							for (t_set_pair::iterator ith = ones.begin(); ith != ones.end(); ++ith)
								merge_set.insert(
									std::pair<
									std::pair<int, int>,
									bool
									>(std::pair<int, int>(ith->first, ith->second), true)
								);
							
						
							for (t_set_pair::iterator ith = oths.begin(); ith != oths.end(); ++ith)
								merge_set.insert(
									std::pair<
									std::pair<int, int>,
									bool
									>(std::pair<int, int>(ith->first, ith->second), false)
								);
							/*for (t_set_triple::iterator itt = merge_set.begin(); itt != merge_set.end(); ++itt)
								std::cout << "(" << (itt->first).first << ", " << (itt->first).second << ", " << itt->second << "), ";
							std::cout << '\n';*/

							/* convert set to vector */
							t_vector_triple merge_vector;
							for (t_set_triple::iterator itt = merge_set.begin(); itt != merge_set.end(); ++itt)
								merge_vector.push_back(*itt);
//								std::cout << "(" << (itt->first).first << ", " << (itt->first).second << ", " << itt->second << "), ";
							/* to test it */
					
#if 0
							std::cout << "Just test: " << merge_vector.size() << " |  ";
							for (t_vector_triple::iterator itv = merge_vector.begin(); itv != merge_vector.end(); ++itv)
								std::cout << "(" << (itv->first).first << ", " << (itv->first).second << ", " << itv->second << "), ";
							std::cout << '\n';
#endif
				
							t_model to_push(n, relation, ortho_left, ortho_right, merge_set, merge_vector);
							models.push_back(to_push);

							//generate_hyper_pseudo_EA(n, merge_set, ortho_left, ortho_right, relation);
 					//	}
					}

				}
			} while (next_perm(1, n - 2, ortho_left));
		}
	}
	else {
		if (x != y) {
			//std::cout << "x = " << x << ", y = " << y << '\n';
			relation[x][y] = 0;
			relation[y][x] = 0;
			get_next_element(x, y, n, relation, models);
			relation[x][y] = 1;
			relation[y][x] = 0;
			get_next_element(x, y, n, relation, models);


			relation[x][y] = 0;
			relation[y][x] = 1;
			get_next_element(x, y, n, relation, models);


		}
	}
}


void generate_hyper_pseudo_EA2(const t_model &model) {
	int min_values[MAX*MAX];

	int model_size = model.v.size();

	// init min values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			min_values[i] = 1;
		}
		else if (x == 0) {
			/* y in 0 + y */
			min_values[i] = 1 << y;
		}
		else if (y == 0) {
			/* x in x + 0 */
			min_values[i] = 1 << x;
		}
		else {
			if (b) {
				/* 1 in x + y */
				min_values[i] = 1 << (model.n - 1);
			}
			else {
				/* {0} is not equal to x + y */
				min_values[i] = 2;
			}
		}
	}


	generate_hyper_pseudo_EA3(model, model_size, min_values);
}

bool check_x_plus_0(const t_model &model, int x, int val) {

	for (int yt = 0; yt < model.n; yt++) {
		// for any yt such that yt + x exists
		if (model.relation[yt][model.ortho_right[x]]) {
			for (int t = 0; t < model.n; t++)
				// if there is t in val-set for which yt + t does not exit
				if (val & (1 << t) && !model.relation[yt][model.ortho_right[t]]) {
					return false;
				}
		}
	}

	return true;
}

bool check_0_plus_y(const t_model &model, int y, int val) {

	for (int xt = 0; xt < model.n; xt++) {
		// for any xt such that y + xt exists
		if (model.relation[y][model.ortho_right[xt]]) {
			for (int t = 0; t < model.n; t++)
				// if there is t in val-set for which t + xt does not exit
				if (val & (1 << t) && !model.relation[t][model.ortho_right[xt]]) {
					return false;
				}
		}
	}

	return true;
}


void generate_hyper_pseudo_EA3(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit) {
	unsigned long int value_generated = 0UL;

	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;


	int var[MAX*MAX];	
	int min_values[MAX*MAX];
	int max_values[MAX*MAX];

	int model_size = model.v.size();

	for (int i = 0; i < model_size; i++) {
		var[i] = 0;
	}

	// init min values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			min_values[i] = 1;
		}
		else if (x == 0) {
			/* y in 0 + y */
			min_values[i] = pow2s[y];
		}
		else if (y == 0) {
			/* x in x + 0 */
			min_values[i] = pow2s[x];
		}
		else {
			if (b) {
				/* 1 in x + y */
				min_values[i] = pow2s[model.n - 1];
			}
			else {
				/* {0} is not equal to x + y */
				min_values[i] = 2;
			}
		}
	}

	// init max values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			max_values[i] = 1;
		}
		else if (x == 0) {
			if (!b) {
				/* 1 not in 0 + y */
				max_values[i] = pow2s[model.n - 1] - 1;
			}
			else
				max_values[i] = pow2s[model.n] - 1;
		}
		else if (y == 0) {
			if (!b) {
				/* 1 not in 0 + y */
				max_values[i] = pow2s[model.n - 1] - 1;
			}
			else
				max_values[i] = pow2s[model.n] - 1;
		}
		else {
			if (!b) {
				/* 1 not in 0 + y */
				max_values[i] = pow2s[model.n - 1] - 1;
			}
			else
				max_values[i] = pow2s[model.n] - 1;
		}
	}

	// init variation to min_values
	for (int i = 0; i < model_size; i++)
		var[i] = min_values[i];
	for (int i = 0; i < min_values_count; i++)
		var[i] = init_min_values[i];

#ifdef DEBUG
	std::cout << "min values: ";
	for (int i =  0; i < model_size; i++)
		std::cout << min_values[i] << ' ';
	std::cout << '\n';

	std::cout << "max values: ";
	for (int i =  0; i < model_size; i++)
		std::cout << max_values[i] << ' ';
	std::cout << '\n';
#endif
	

	int next_idx;
	do {
#ifdef DEBUG
		for (int i =  0; i < model_size; i++)
			std::cout << var[i] << ' ';

		std::cout << '\n';
#endif
		bool is_ok = true;
		next_idx = model_size - 1;
		for (int i =  0; is_ok && (i < model_size); i++) {
			//bool b = model.v[i].second;
			int x = model.v[i].first.first,
			    y = model.v[i].first.second; 	
			
			if (x == 0 && y == 0) {
			}
			else if (x == 0) {

				if (!(var[i] & pow2s[y]) || !check_0_plus_y(model, y, var[i])) {
					is_ok = false;
					next_idx = i;
					break;
				}
/*
				bool exist_next = true;
				while (exist_next && (!(var[i] & (1 << y)) || !check_0_plus_y(model, y, var[i]))) {
					exist_next = get_next_variation(min_values[i], max_values[i], i, i, var);
					
				}
				if (!exist_next) {
					is_ok = false;
					break;
				}
*/

			}
			else if (y == 0) {
				if (!(var[i] &	pow2s[x]) || !check_x_plus_0(model, x, var[i])) {
					is_ok = false;
					next_idx = i;
					break;
				}
#if 0
				bool exist_next = true;
				while (exist_next && (!(var[i] & (1 << x)) || !check_x_plus_0(model, x, var[i]))) {
					exist_next = get_next_variation(min_values[i], max_values[i], i, i, var);
					
				}
				if (!exist_next) {
					is_ok = false;
					break;
				}
#endif
			}
			else {
			}
			hypea[x][y] = var[i];
		}
		if (is_ok) {
			value_generated++;
			if (value_generated == generated_limit) {
				// dump state
				for (int i = 0; i < model_size;i++)
					std::cerr << var[i] << ',';
				std::cerr << '\n';
				value_generated = 0UL;
			}
			if (is_associative(model.n, hypea)) {
				/*for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						std::cout << hypea[i][j] << ", ";
					std::cout << '\n';
				}*/
				disp_hypea(model.n, hypea);
				std::cout << "---\n";
			}
		}
		else {
			// init (next_idx + 1) .. (model_size - 1) values 
			//	in var to min_values[idx]
			for (int k = next_idx + 1; k < model_size; k++)
				var[k] = min_values[k];
		}
	} while (get_next_variation(min_values, max_values, 0, next_idx, var));
	

}

bool is_summable_elem_elem(const t_model &model, int e1, int e2) {
	return model.relation[e1][model.ortho_right[e2]];
}

bool is_summable_set_elem(const t_model &model, int s, int e) {
	for (int i = 0; i < model.n; i++)
		if ((s & pow2s[i]) && !is_summable_elem_elem(model, i, e))
			return false;

	return true;
}

bool is_summable_elem_set(const t_model &model, int e, int s) {
	for (int i = 0; i < model.n; i++)
		if ((s & pow2s[i]) && !is_summable_elem_elem(model, e, i))
			return false;

	return true;
}


void generate_hyper_pseudo_EA4(const t_model &model, unsigned long int generated_limit) {
	unsigned long int value_generated = 0UL;

	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;

	std::vector<std::vector<int> > possible_values;

	int var[MAX*MAX];	
	int min_values[MAX*MAX];
	int max_values[MAX*MAX];

	int model_size = model.v.size();

	for (int i = 0; i < model_size; i++) {
		var[i] = 0;
	}

	// init possible values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			std::vector<int> pom;
			pom.push_back(1);
			possible_values.push_back(pom);

/*
			min_values[i] = 1;
*/
		}
		else if (x == 0) {
			/* y in 0 + y */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[y]) && check_0_plus_y(model, y, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else if (y == 0) {
			/* x in x + 0 */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[x]) && check_x_plus_0(model, x, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else {
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			if (b) {
				for (int i = start; i <= end; i++)
					pom.push_back(i);
			}
			else {
				for (int i = start; i <= end; i++) {
					bool ok = true;
					// if set i is summable with some non-zero element z
					// from right, then y + z must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_set_elem(model, i, j)) {
							if (!is_summable_elem_elem(model, y, j)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					// if set i is summable with some non-zero element z
					// from left, then z + x must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_elem_set(model, j, i)) {
							if (!is_summable_elem_elem(model, j, x)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					if (ok) 
						pom.push_back(i);

				}
			}
			possible_values.push_back(pom);
		}
	}

	// init min, max values
	for (int i = 0; i < model_size; i++) {
		min_values[i] = 0;
		max_values[i] = possible_values[i].size() - 1;

		if (max_values < 0) {
			std::cerr << "No models!";
			return ;
		}
	}

	// Dump possible values to std::cerr
	std::cerr << "Possible values: \n";
	for (int i = 0; i < model_size; i++) {
		std::cerr << "\t(" << model.v[i].first.first << ", " << model.v[i].first.second << "): ";
		for (int j = 0; j <= max_values[i]; j++)
			std::cerr << possible_values[i][j] << ", ";
		std::cerr << '\n';
	}
	//
		
	do {
		for (int i =  0; i < model_size; i++) {
			int x = model.v[i].first.first,
			    y = model.v[i].first.second; 	
			
			hypea[x][y] = possible_values[i][var[i]];
		}

		value_generated++;
		if (value_generated == generated_limit) {
			// dump state
			for (int i = 0; i < model_size;i++)
				std::cerr << var[i] << ',';
			std::cerr << '\n';
			std::cerr << '(';
			for (int i = 0; i < model_size;i++)
				std::cerr << possible_values[i][var[i]] << ',';
			std::cerr << ")\n";

			value_generated = 0UL;
		}
		if (is_associative(model.n, hypea)) {
			/*for (int i = 0; i < n; i++) {
				for (int j = 0; j < n; j++)
					std::cout << hypea[i][j] << ", ";
				std::cout << '\n';
			}*/
			disp_hypea(model.n, hypea);
			std::cout << "---\n";
		}
	} while (get_next_variation(min_values, max_values, 0, model_size - 1, var));
	

}

void generate_hyper_pseudo_EA5(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit) {
	unsigned long int value_generated = 0UL;

	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;

	std::vector<std::vector<int> > possible_values;

	int var[MAX*MAX];	
	int min_values[MAX*MAX];
	int max_values[MAX*MAX];

	int model_size = model.v.size();

	for (int i = 0; i < model_size; i++) {
		var[i] = 0;
	}

	// init possible values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			std::vector<int> pom;
			pom.push_back(1);
			possible_values.push_back(pom);

/*
			min_values[i] = 1;
*/
		}
		else if (x == 0) {
			/* y in 0 + y */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[y]) && check_0_plus_y(model, y, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else if (y == 0) {
			/* x in x + 0 */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[x]) && check_x_plus_0(model, x, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else {
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			if (b) {
				for (int i = start; i <= end; i++)
					pom.push_back(i);
			}
			else {
				for (int i = start; i <= end; i++) {
					bool ok = true;
					// if set i is summable with some non-zero element z
					// from right, then y + z must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_set_elem(model, i, j)) {
							if (!is_summable_elem_elem(model, y, j)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					// if set i is summable with some non-zero element z
					// from left, then z + x must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_elem_set(model, j, i)) {
							if (!is_summable_elem_elem(model, j, x)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					if (ok) 
						pom.push_back(i);

				}
			}
			possible_values.push_back(pom);
		}
	}

	// init min, max values
	for (int i = 0; i < model_size; i++) {
		min_values[i] = 0;
		max_values[i] = possible_values[i].size() - 1;

		if (max_values < 0) {
			std::cerr << "No models!";
			return ;
		}
	}
	// Dump possible values to std::cerr
	std::cerr << "Possible values: \n";
	for (int i = 0; i < model_size; i++) {
		std::cerr << "\t(" << model.v[i].first.first << ", " << model.v[i].first.second << "): ";
		for (int j = 0; j <= max_values[i]; j++)
			std::cerr << possible_values[i][j] << ", ";
		std::cerr << '\n';
	}


	// update starting state
	for (int i = 0; i < min_values_count; i++) {
		var[i] = init_min_values[i];
	}

	// print starting state
	std::cerr << "Starting from: ";
	for (int i = 0; i < min_values_count; i++) {
		std::cerr << init_min_values[i] << ",";
	}
	std::cerr << '\n';
	std::cerr << '(';
	for (int i = 0; i < min_values_count;i++)
		std::cerr << possible_values[i][init_min_values[i]] << ',';
	std::cerr << ")\n";
	

	do {
		for (int i =  0; i < model_size; i++) {
			int x = model.v[i].first.first,
			    y = model.v[i].first.second; 	
			
			hypea[x][y] = possible_values[i][var[i]];
		}

		value_generated++;
		if (value_generated == generated_limit) {
			// dump state
			for (int i = 0; i < model_size;i++)
				std::cerr << var[i] << ',';
			std::cerr << '\n';
			std::cerr << '(';
			for (int i = 0; i < model_size;i++)
				std::cerr << possible_values[i][var[i]] << ',';
			std::cerr << ")\n";

			value_generated = 0UL;
		}
		if (is_associative(model.n, hypea)) {
			/*for (int i = 0; i < n; i++) {
				for (int j = 0; j < n; j++)
					std::cout << hypea[i][j] << ", ";
				std::cout << '\n';
			}*/
			disp_hypea(model.n, hypea);
			std::cout << "---\n";
		}
	} while (get_next_variation(min_values, max_values, 0, model_size - 1, var));
	

}

void generate_hyper_pseudo_EA5_just(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit) {
	unsigned long int value_generated = 0UL;

	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;

	std::vector<std::vector<int> > possible_values;

	int var[MAX*MAX];	
	int min_values[MAX*MAX];
	int max_values[MAX*MAX];

	int model_size = model.v.size();

	for (int i = 0; i < model_size; i++) {
		var[i] = 0;
	}

	// init possible values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			std::vector<int> pom;
			pom.push_back(1);
			possible_values.push_back(pom);

/*
			min_values[i] = 1;
*/
		}
		else if (x == 0) {
			/* y in 0 + y */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[y]) && check_0_plus_y(model, y, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else if (y == 0) {
			/* x in x + 0 */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[x]) && check_x_plus_0(model, x, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else {
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			if (b) {
				for (int i = start; i <= end; i++)
					pom.push_back(i);
			}
			else {
				for (int i = start; i <= end; i++) {
					bool ok = true;
					// if set i is summable with some non-zero element z
					// from right, then y + z must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_set_elem(model, i, j)) {
							if (!is_summable_elem_elem(model, y, j)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					// if set i is summable with some non-zero element z
					// from left, then z + x must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_elem_set(model, j, i)) {
							if (!is_summable_elem_elem(model, j, x)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					if (ok) 
						pom.push_back(i);

				}
			}
			possible_values.push_back(pom);
		}
	}

	// init min, max values
	for (int i = 0; i < model_size; i++) {
		min_values[i] = 0;
		max_values[i] = possible_values[i].size() - 1;

		if (max_values < 0) {
			std::cerr << "No models!";
			return ;
		}
	}
	// Dump possible values to std::cerr
	std::cerr << "Possible values: \n";
	for (int i = 0; i < model_size; i++) {
		std::cerr << "\t(" << model.v[i].first.first << ", " << model.v[i].first.second << "): ";
		for (int j = 0; j <= max_values[i]; j++)
			std::cerr << possible_values[i][j] << ", ";
		std::cerr << '\n';
	}


	// update starting state
	for (int i = 0; i < min_values_count; i++) {
		var[i] = init_min_values[i];
	}

	// print starting state
	std::cerr << "Starting from: ";
	for (int i = 0; i < min_values_count; i++) {
		std::cerr << init_min_values[i] << ",";
	}
	std::cerr << '\n';
	std::cerr << '(';
	for (int i = 0; i < min_values_count;i++)
		std::cerr << possible_values[i][init_min_values[i]] << ',';
	std::cerr << ")\n";
	

	do {
		for (int i =  0; i < model_size; i++) {
			int x = model.v[i].first.first,
			    y = model.v[i].first.second; 	
			
			hypea[x][y] = possible_values[i][var[i]];
		}

		value_generated++;
		if (value_generated == generated_limit) {
			// dump state
			for (int i = 0; i < model_size;i++)
				std::cerr << var[i] << ',';
			std::cerr << '\n';
			std::cerr << '(';
			for (int i = 0; i < model_size;i++)
				std::cerr << possible_values[i][var[i]] << ',';
			std::cerr << ")\n";

			value_generated = 0UL;
		}
		if (is_associative(model.n, hypea)) {
			/*for (int i = 0; i < n; i++) {
				for (int j = 0; j < n; j++)
					std::cout << hypea[i][j] << ", ";
				std::cout << '\n';
			}*/
			disp_hypea(model.n, hypea);
			std::cout << "---\n";
		}
	} while (get_next_variation(min_values, max_values, min_values_count, model_size - 1, var));
	

}

void generate_hyper_pseudo_EA6(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit) {
	unsigned long int value_generated = 0UL;

	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;

	std::vector<std::vector<int> > possible_values;

	int var[MAX*MAX];	
	int min_values[MAX*MAX];
	int max_values[MAX*MAX];

	int model_size = model.v.size();

	for (int i = 0; i < model_size; i++) {
		var[i] = 0;
	}

	// init possible values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			std::vector<int> pom;
			pom.push_back(1);
			possible_values.push_back(pom);

/*
			min_values[i] = 1;
*/
		}
		else if (x == 0) {
			/* y in 0 + y */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[y]) && check_0_plus_y(model, y, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else if (y == 0) {
			/* x in x + 0 */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[x]) && check_x_plus_0(model, x, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else {
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			if (b) {
				for (int i = start; i <= end; i++)
					pom.push_back(i);
			}
			else {
				for (int i = start; i <= end; i++) {
					bool ok = true;
					// if set i is summable with some non-zero element z
					// from right, then y + z must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_set_elem(model, i, j)) {
							if (!is_summable_elem_elem(model, y, j)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					// if set i is summable with some non-zero element z
					// from left, then z + x must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_elem_set(model, j, i)) {
							if (!is_summable_elem_elem(model, j, x)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					if (ok) 
						pom.push_back(i);

				}
			}
			possible_values.push_back(pom);
		}
	}

	// init min, max values
	for (int i = 0; i < model_size; i++) {
		min_values[i] = 0;
		max_values[i] = possible_values[i].size() - 1;

		if (max_values < 0) {
			std::cerr << "No models!";
			return ;
		}
	}
	// Dump possible values to std::cerr
	std::cerr << "Possible values: \n";
	for (int i = 0; i < model_size; i++) {
		std::cerr << "\t(" << model.v[i].first.first << ", " << model.v[i].first.second << "): ";
		for (int j = 0; j <= max_values[i]; j++)
			std::cerr << possible_values[i][j] << ", ";
		std::cerr << '\n';
	}


	if (min_values_count) {
		// update starting state
		for (int i = 0; i < min_values_count; i++) {
			var[i] = init_min_values[i];
		}

		// print starting state
		std::cerr << "Starting from: ";
		for (int i = 0; i < min_values_count; i++) {
			std::cerr << init_min_values[i] << ",";
		}
		std::cerr << '\n';
		std::cerr << '(';
		for (int i = 0; i < min_values_count;i++)
			std::cerr << possible_values[i][init_min_values[i]] << ',';
		std::cerr << ")\n";
	}
	

	bool b_ok; 
	int last_idx; 
	do {
		b_ok = true;
		last_idx = model_size - 1;

		for (int i =  0; b_ok && (i < model_size); i++) {
			int x = model.v[i].first.first,
			    y = model.v[i].first.second; 	
			
			if (x == 0 && y) {
				/* */
				int tested_value = possible_values[i][var[i]];
				for (int k = 0; k < y; k++)
					if ((pow2s[k] & tested_value) && ((hypea[0][k] & tested_value) != hypea[0][k])) {
						b_ok = false;
						last_idx = i;
						break;
					}
			}
			else if (y == 0 && x) {
				/* */
				int tested_value = possible_values[i][var[i]];
				for (int k = 0; k < x; k++)
					if ((pow2s[k] & tested_value) && ((hypea[k][0] & tested_value) != hypea[k][0])) {
						b_ok = false;
						last_idx = i;
						break;
					}
			}

			if (!b_ok)
				break;

			hypea[x][y] = possible_values[i][var[i]];
		}

		if (b_ok) {
			value_generated++;
			if (value_generated == generated_limit) {
				// dump state
				for (int i = 0; i < model_size;i++)
					std::cerr << var[i] << ',';
				std::cerr << '\n';
				std::cerr << '(';
				for (int i = 0; i < model_size;i++)
					std::cerr << possible_values[i][var[i]] << ',';
				std::cerr << ")\n";

				value_generated = 0UL;
			}
			if (is_associative(model.n, hypea)) {
				/*for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						std::cout << hypea[i][j] << ", ";
					std::cout << '\n';
				}*/
				disp_hypea(model.n, hypea);
				std::cout << "---\n";
			}
		}
		else {
			for (int k = last_idx + 1; k < model_size; k++)
				var[k] = 0;
		}
	} while (get_next_variation(min_values, max_values, 0, last_idx, var));
	

}

void generate_hyper_pseudo_EA6_just(const t_model &model, int min_values_count, int init_min_values[MAX*MAX], unsigned long int generated_limit) {
	unsigned long int value_generated = 0UL;

	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;

	std::vector<std::vector<int> > possible_values;

	int var[MAX*MAX];	
	int min_values[MAX*MAX];
	int max_values[MAX*MAX];

	int model_size = model.v.size();

	for (int i = 0; i < model_size; i++) {
		var[i] = 0;
	}

	// init possible values
	for (int i = 0; i < model_size; i++) {
		bool b = model.v[i].second;
		int x = model.v[i].first.first,
		    y = model.v[i].first.second; 	

		if (x == 0 && y == 0) {
			/* 0 + 0 = {0} */
			std::vector<int> pom;
			pom.push_back(1);
			possible_values.push_back(pom);

/*
			min_values[i] = 1;
*/
		}
		else if (x == 0) {
			/* y in 0 + y */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[y]) && check_0_plus_y(model, y, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else if (y == 0) {
			/* x in x + 0 */
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			for (int i = start; i <= end; i++) {
				if ((i & pow2s[x]) && check_x_plus_0(model, x, i)) {
					pom.push_back(i);
				}
			}
			possible_values.push_back(pom);
		}
		else {
			std::vector<int> pom;
			int start, end;
			if (b) {
				start = pow2s[model.n - 1];
				end = pow2s[model.n] - 1;	
			}
			else {
				start = 2;
				end = pow2s[model.n - 1] - 1;
			}

			if (b) {
				for (int i = start; i <= end; i++)
					pom.push_back(i);
			}
			else {
				for (int i = start; i <= end; i++) {
					bool ok = true;
					// if set i is summable with some non-zero element z
					// from right, then y + z must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_set_elem(model, i, j)) {
							if (!is_summable_elem_elem(model, y, j)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					// if set i is summable with some non-zero element z
					// from left, then z + x must exist
					for (int j = 1; j < model.n; j++) {
						if (is_summable_elem_set(model, j, i)) {
							if (!is_summable_elem_elem(model, j, x)) {
								ok = false;
								break;
							}
						}
						
					}
					if (!ok)
						continue;

					if (ok) 
						pom.push_back(i);

				}
			}
			possible_values.push_back(pom);
		}
	}

	// init min, max values
	for (int i = 0; i < model_size; i++) {
		min_values[i] = 0;
		max_values[i] = possible_values[i].size() - 1;

		if (max_values < 0) {
			std::cerr << "No models!";
			return ;
		}
	}
	// Dump possible values to std::cerr
	std::cerr << "Possible values: \n";
	for (int i = 0; i < model_size; i++) {
		std::cerr << "\t(" << model.v[i].first.first << ", " << model.v[i].first.second << "): ";
		for (int j = 0; j <= max_values[i]; j++)
			std::cerr << possible_values[i][j] << ", ";
		std::cerr << '\n';
	}


	// update starting state
	for (int i = 0; i < min_values_count; i++) {
		var[i] = init_min_values[i];
	}

	// print starting state
	std::cerr << "Starting from: ";
	for (int i = 0; i < min_values_count; i++) {
		std::cerr << init_min_values[i] << ",";
	}
	std::cerr << '\n';
	std::cerr << '(';
	for (int i = 0; i < min_values_count;i++)
		std::cerr << possible_values[i][init_min_values[i]] << ',';
	std::cerr << ")\n";
	

	bool b_ok; 
	int last_idx; 
	do {
		b_ok = true;
		last_idx = model_size - 1;
		for (int i =  0; b_ok && (i < model_size); i++) {
			int x = model.v[i].first.first,
			    y = model.v[i].first.second; 	
			

			if (x == 0 && y) {
				/* */
				int tested_value = possible_values[i][var[i]];
				for (int k = 0; k < y; k++)
					if ((pow2s[k] & tested_value) && ((hypea[0][k] & tested_value) != hypea[0][k])) {
						b_ok = false;
						last_idx = i;
						break;
					}
			}
			else if (y == 0 && x) {
				/* */
				int tested_value = possible_values[i][var[i]];
				for (int k = 0; k < x; k++)
					if ((pow2s[k] & tested_value) && ((hypea[k][0] & tested_value) != hypea[k][0])) {
						b_ok = false;
						last_idx = i;
						break;
					}
			}

			if (!b_ok)
				break;
			
			hypea[x][y] = possible_values[i][var[i]];
		}

		if (b_ok) {
			value_generated++;
			if (value_generated == generated_limit) {
				// dump state
				for (int i = 0; i < model_size;i++)
					std::cerr << var[i] << ',';
				std::cerr << '\n';
				std::cerr << '(';
				for (int i = 0; i < model_size;i++)
					std::cerr << possible_values[i][var[i]] << ',';
				std::cerr << ")\n";

				value_generated = 0UL;
			}
			if (is_associative(model.n, hypea)) {
				/*for (int i = 0; i < n; i++) {
					for (int j = 0; j < n; j++)
						std::cout << hypea[i][j] << ", ";
					std::cout << '\n';
				}*/
				disp_hypea(model.n, hypea);
				std::cout << "---\n";
			}
		}
		else {
			if (last_idx < min_values_count)
				return;

			for (int k = last_idx + 1; k < model_size -1; k++)
				var[k] = 0;
		}
	} while (get_next_variation(min_values, max_values, min_values_count, last_idx, var));
	

}

#if 0
void generate_hyper_pseudo_EA(const t_model &model, int from_length, int (&from_state)[MAX*MAX], int to_length, int (&to_state)[MAX*MAX]) {
	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;
	int var[MAX*MAX];	

	for (int i = 0; i < from_length; i++)
		var[i] = from_state[i];

	for (int j = from_length; j < model.v.size(); j++) {
		bool b = model.v[j].second;

		if (!b)
			var[j] = 2;
		else 
			var[j] = 1 << (n-1);
	}

	do {
		for (int i = 0; i < model.v.size(); i++) {
			int x = model.v[i].first.first,
				y = model.v[i].first.second;
			bool b = model.v[i].second;

			if (x == 0 && y == 0) {
				if (var[i] > to_state[i])
					return;			
				
			}
		}
		
	}
		

	



	generate_hyper_pseudo_EA_rec(model.t.begin(), hypea, model.n, model.t, model.ortho_left, model.ortho_right, model.relation);
}
#endif

void skip_ws(char **from) {
	char *pom = *from;

	while (*pom && (*pom == '\t' || *pom == ' '))
		pom++;

	*from = pom;
}

bool get_num(char **pom, int &value) {
	int val = 0;
	bool at_least_one = false;
	char *pomtmp = *pom;

	while (*pomtmp && (*pomtmp >='0' && *pomtmp <= '9')) {
		val = 10 *val + (*pomtmp - '0');
		at_least_one = true;
		pomtmp++;
	}

	if (at_least_one) {
		value = val;
		*pom = pomtmp;
	}
	

	return at_least_one;
}

void generate_hyper_pseudo_EA(const t_model &model) {
	int hypea[MAX][MAX];
	for (int i = 0; i < model.n; i++)
		for (int j = 0; j < model.n; j++)
			hypea[i][j] = 0;
	
	generate_hyper_pseudo_EA_rec(model.t.begin(), hypea, model.n, model.t, model.ortho_left, model.ortho_right, model.relation);
}

void disp_set_triple(t_set_triple &triple) {

	for (t_set_triple::iterator itt = triple.begin(); itt != triple.end(); ++itt)
	std::cout << "(" << (itt->first).first << ", " << (itt->first).second << ", " << itt->second << "), ";
	std::cout << '\n';
}

void print_model(const t_model &model) {

	for (int i = 0; i < model.n; i++) {
		for (int j = 0; j < model.n; j++) {
			std::cerr << model.relation[i][j];
		}
		std::cerr << '\n';
		
	}
	std::cerr << "-\n";

	std::cerr << "Ortho_left : ";
	for (int i = 0; i < model.n; i++)
		std::cerr << model.ortho_left[i];
	std::cerr << '\n';

	std::cerr << "Ortho_right: ";
	for (int i = 0; i < model.n; i++)
		std::cerr << model.ortho_right[i];
	std::cerr << '\n';

	std::cerr << "Ones: ";
	for (std::vector<t_triple>::const_iterator it = model.v.begin(); it != model.v.end(); ++it) {
		if (it -> second) {
			std::cerr << "(" << (it->first).first << ", " << (it->first).second << "), ";
		}
	}
	std::cerr << '\n';

	std::cerr << "Oths: ";
	for (t_vector_triple::const_iterator it = model.v.begin(); it != model.v.end(); ++it) {
		if (!(it -> second)) {
			std::cerr << "(" << (it->first).first << ", " << (it->first).second << "), ";
		}
	}
	std::cerr << '\n';
	std::cerr << "---\n";
}

int main(int argc, char **argv) {
	if (argc < 2) {
		std::cout << "Usage: " << argv[0] << " <n> [<model_type_no>]\n";
		std::cout << "   n - number of elements "<< MAX << ">=n>=3\n";
		std::cout << "   model_type_no - generate hyper pseudo effect algebras according to model_type_no\n";
		std::cout << " If called with just one parameter, it shows just possible model types and their count.\n";
		return 0;
	}
	int n = atoi(argv[1]);

	if (n < 3 || n > MAX) {
		std::cerr << "n is out of range, must be 3 <= n << " << MAX << ".\n";
		return 0;
	}

	std::vector<t_model> models;	

	generate_all_bounded_relations(n, models);

	int max_models;

	max_models = models.size();

	if (argc < 3) {
		for (int i = 0; i < max_models; i++)
			print_model(models[i]);

		std::cerr << "For n = " << n << " there are " << max_models << " model types.\n";

		return 0;
	}


	int model = atoi(argv[2]);

	if (model < 1 || model > max_models) {
		std::cerr << "model_no is out of range, must be 1 <= model_no <= " << max_models << ".\n";
		return 0;
	}

	if (argc < 4) {
		int init_min[MAX*MAX];
		print_model(models[model - 1]);
		generate_hyper_pseudo_EA6(models[model - 1],0,init_min);
	}
	else {
		/* parse init_state */
		int init_min[MAX*MAX];

		char *pom = argv[3];
		int idx = 0;
		int val;

		while (*pom) {
			skip_ws(&pom);
			if (get_num(&pom, val)) {
//				std::cout << val << "M";
				init_min[idx] = val;
				idx++;
			}
			else {
				std::cerr << "Expecting number value\n";
				return 0;
			}

			skip_ws(&pom);
			if (*pom) {
				if (*pom !=',') {
					std::cerr << "Missing comma delimiter\n";
					return 0;
				}
				else
					pom++;
			}
		}

		int just = 0;
		if (argc < 5) {}
		else {
			just = atoi(argv[4]);
			if (just != 0 && just != 1) {
				std::cerr << "Expecting values ...\n \t0 (all models from) and\n\t 1 (just models from having starting state).\n";
				return 0;
			}
		}
			

		std::cerr << "From: ";
		{
			int i = 0;
			const int ii = idx - 1;
			while (i < ii) {
				std::cerr << init_min[i] << ",";
				i++;
			}
			
			std::cerr << init_min[i];
			std::cerr << '\n';
		}

		if (just)
			generate_hyper_pseudo_EA6_just(models[model - 1], idx, init_min);
		else
			generate_hyper_pseudo_EA6(models[model - 1], idx, init_min);

#if 0
		std::cout << "VALUES: ";
		for (int i = 0; i < idx; i++)
			std::cout << init_min[i] << " ";
		std::cout << '\n';
#endif
	}
#if 0
	int perm[5] = { 1,2,3,4,5 };
	do {
		for (int i = 0; i < 5; i++)
			std::cout << perm[i];
		std::cout << '\n';
		
	} while (next_perm(1, 3, perm));
#endif
#if 0
	if (is_less_relation(2, rel2, rel1))
		std::cout << "Ok\n";

	
	hyper_pseudo_EA<2> my_hpea;

	generate_all_hyper_pseudo_EA(0, 0, my_hpea);	
#endif
	return 0;
}
